﻿/* MovinGifCreator
 * 
 * COPYLEFT (C') CharMY 2014 Yanagihara Masami
 * 
 * 
 * This program is free software: you can redistribute it and/or modify   *
 * it under the terms of the GNU General Public License as published by   *
 * the Free Software Foundation, either version 3 of the License, or      *
 * (at your option) any later version.                                    *
 *                                                                        *
 * This program is distributed in the hope that it will be useful,        *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 * 
 * 
 */

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using CharMY;
using CharMY.Analysis;
using CharMY.Analysis.Interfaces;
using CharMY.Collections.Generic;
using CharMY.Drawing;
using CharMY.IO;
using CharMY.Resources;
using CharMY.Text;
using CharMY.Windows.Forms;

namespace CharMY.MovinGif
{
    using Context = ContextStripItemAttribute;
    using GifSegments = GridCollection<GifSegment, object>;
    using Menu = MenuStripItemAttribute;
    using Status = StatusStripItemAttribute;
    using Tool = ToolStripItemAttribute;
    internal class MovinGifMenuSet : MenuSetCore<MatrixGridView>
    {
        #region メニューGUIの自動生成

        #region I/O

        [Menu(2, indexFile, Shortcut = Keys.Control | Keys.O)]
        public void LoadImageFiles()
        {
            var files = Files.ImagesIO().GetFiles();
            if (files.Key == null)
                return;

            this.FromGifImage(files.Value
                .ZipValue(f => files.Key.Load(f)),
                gss =>
                {
                    _data.Records = gss;
                    this.Content.RefleshAllCell();
                });
        }

        [Menu(3, indexFile, Shortcut = Keys.Control | Keys.S)]
        public void SaveGifFile()
        {
            var io = Files.GifData();
            var file = io.GetFile(FileOpenOptions.SaveDefault);
            if (file == null)
                return;

            ProgressControls.BeginHeavyComputing(pc =>
            {
                io.Save(this.Create(pc), file);
            });
        }

        [Menu(4, indexFile)]
        public void SaveImages()
        {
            var pair = Files.ImagesIO().GetFile(FileOpenOptions.SaveDefault);
            var io = pair.Key;
            var default_file = pair.Value;
            int index = 0;
            foreach (var img in _data.Records)
            {
                ++index;
                if (img.Image != null)
                {
                    io.Save(img.Image, default_file.ChangeName((name, ext) => name + index.ToString("D3") + ext));
                }
            }
        }

        #endregion

        #region Edit

        const int indexEdit = indexFile + 0x10;
        const int __i_paste_as = indexSeparator | 0x5;

        [Context(4, indexContextMain)]
        [Menu(4 | indexSeparator, indexEdit, "Edit", Shortcut = Keys.Delete)]
        public void Delete()
        {
            var selection = this.Content.SelectionIndexRange;
            if (0 <= selection.Top)
            {
                _data.Records = _data.Records
                    .Where((_, i) => i < selection.Top || selection.Bottom <= i)
                    .ToArray();

                this.Content.RefleshAllCell();
            }
        }

        [Context(1, indexContextMain)]
        [Menu(1 | indexSeparator, indexEdit, Shortcut = Keys.Control | Keys.X)]
        public void Cut()
        {
            this.Copy();
            this.Delete();
        }

        [Context(2, indexContextMain)]
        [Menu(2 | indexSeparator, indexEdit, Shortcut = Keys.Control | Keys.C)]
        public void Copy()
        {
            Image image = null;

            var io = Files.HtmlTable();
            var matrix = this.Content.Items;
            var range = this.Content.SelectionIndexRange;
            var range_r = Enumerable.Range(range.Top, range.Height);
            var range_c = Enumerable.Range(range.Left, range.Width);
            var text = StaticStringBuilder.Take();
            var table_writer = io.MakeWriter(text.ToWriter());
            table_writer
                .WriteTable(range_r
                    .Select(r => range_c
                        .Select(c =>
                        {
                            if (c < 0)
                                return r.ToString();

                            if (c == 0)
                            {
                                image = image ?? (matrix.GetItem(c, r) as Image);
                                return "<image>";
                            }

                            return matrix.GetItemText(c, r);
                        })
                    )
                );

            try
            {
                Clipboard.Clear();
                Clipboard.SetText(text.ToString(), io.Extensions.Contains("csv") ? TextDataFormat.CommaSeparatedValue : TextDataFormat.Text);
                if (image != null)
                    Clipboard.SetImage(image);
            }
            catch (Exception e) { Console.WriteLine(e.Message); }
        }

        [Context(3, indexContextMain)]
        [Menu(3 | indexSeparator, indexEdit, Shortcut = Keys.Control | Keys.V)]
        public void Paste()
        {
            if (Clipboard.ContainsImage())
            {
                this.FromGifImage(null, Clipboard.GetImage(), this.PasteAdd);
            }
            else if (Clipboard.ContainsFileDropList())
            {
                var files = Clipboard.GetFileDropList()
                    .Cast<string>()
                    .Select(p => p.ToFile())
                    .Where(f => f != null && f.Exists)
                    .ZipValue(f => (ImageData)Image.FromFile(f.FullName))
                    .ToArray();

                if (files.Length > 0)
                {
                    this.FromGifImage(files, this.PasteAdd);
                }
            }
        }
        private void PasteAdd(GifSegment[] gss)
        {
            var selection = this.Content.SelectionIndexRange;
            int next_index = selection.Bottom;
            if (next_index < 0)
                next_index = 0x10000;

            {
                var data = _data.Records.ToArrayCast();
                _data.Records = data
                    .Take(next_index)
                    .Concat(gss)
                    .Concat(data.Skip(next_index));

                this.Content.RefleshAllCell();
            }
        }


        [Menu(1 | (indexSeparator * 2), indexEdit)]
        public void SetAllIntervals()
        {
            var vi = WindowForms.ValueInput(0.01M, 60M, 2);
            var opt = vi.ShowDialogSelectOrDefault(
                _data.Records.Select(gs => gs.Interval / 1000M).FirstOrDefault(), -1M, this.PageInfo.Window,
                "Interval Time [seconds]");
            if (opt >= 0.01M)
            {
                for (int r = 0 ; r < base.Content.Items.RowCount ; ++r)
                    base.Content.Items.SetItem(2, r, (opt * 1000).ToInt());
                base.Content.RefleshAllCell();
            }
        }
        [Menu(2 | (indexSeparator * 2), indexEdit)]
        public void SetAllDisposalMethods()
        {
            var vi = WindowForms.ValueInput(typeof(GifData.DisposalMethod).GetEnumNames());
            var opt = vi.ShowDialogSelectOrDefault(
                _data.Records.Select(gs => gs.DisposalMothod.ToString()).FirstOrDefault(), null, this.PageInfo.Window);
            if (opt != null)
            {
                for (int r = 0 ; r < base.Content.Items.RowCount ; ++r)
                    base.Content.Items.SetItem(3, r, opt);
                base.Content.RefleshAllCell();
            }
        }
        [Menu(3 | (indexSeparator * 2), indexEdit)]
        public void SetTransparentColor()
        {
            this.Background();
            _trans.Value = true;
        }

        private readonly Queue<KeyValuePair<DateTime, Bitmap>> _queue = new Queue<KeyValuePair<DateTime, Bitmap>>();
        protected sealed override bool CameraTaken(Bitmap bmp)
        {
            if (base.CameraTakeVideo.Value)
            {
                _queue.Enqueue(bmp.Cache().WithKey(DateTime.Now));
                return true;
            }
            return false;
        }
        protected sealed override void CameraTakeVideoFinished()
        {
            var now = DateTime.Now;
            if (_queue.Count > 0)
            {
                var data1 = _queue.ToArray();
                _queue.Clear();

                var data2 = data1
                    .ZipOffset(0, 1)
                    .Select(offset => new GifSegment()
                    {
                        Image = offset[0].Value,
                        Interval = (int)(offset[1].Key - offset[0].Key)
                            .TotalMilliseconds,
                        DisposalMothod = GifData.DisposalMethod.Fill

                    })
                    .ConcatItem(new GifSegment()
                    {
                        Image = data1[data1.Length - 1].Value,
                        Interval = (int)(now - data1[data1.Length - 1].Key)
                            .TotalMilliseconds,
                        DisposalMothod = GifData.DisposalMethod.Fill

                    })
                    .ToArray();
                this.PageInfo.Window.AddPage(page =>
                {
                    var menu = new MovinGifMenuSet(page);
                    menu._data.Records = data2;
                    return menu;
                });
            }
        }

        const int indexCreateTool = indexToolCamera + 2;
        [Tool(1, indexCreateTool, BaseText = "Gif Creator", BaseStrip = StripPlace.ToolTop)]
        public void CreateGifImage()
        {
            ProgressControls.BeginHeavyComputing(pc =>
            {
                var data = this.Create(pc);
                var iv = data
                    .ToImage()
                    .ImageView();
                iv.Text = data.Data.Count().ToString("N0") + " byte";
                iv.ShowDialogLocated(this.PageInfo.Window);
            });
        }

        private readonly ValueWatcher<bool>
            _common = new ValueWatcher<bool>(true),
            _optim = new ValueWatcher<bool>(true),
            _trans = new ValueWatcher<bool>(true),
            _backc = new ValueWatcher<bool>(false);

        [Tool(7, indexCreateTool)]
        [Menu(7 | (indexSeparator * 2), indexEdit)]
        public ValueWatcher<bool> CommonColorPalette { get { return _common; } }
        [Tool(8, indexCreateTool)]
        public ValueWatcher<bool> Optimization { get { return _optim; } }
        [Tool(9, indexCreateTool)]
        [Menu(9 | (indexSeparator * 2), indexEdit)]
        public ValueWatcher<bool> Transparent { get { return _trans; } }
        [Tool(10, indexCreateTool)]
        [Menu(10 | (indexSeparator * 2), indexEdit)]
        public ValueWatcher<bool> UseBackColor { get { return _backc; } }

        [Tool(2, indexCreateTool, Text = "  Loop =")]
        public void Loop() { this.LoopCount.Value = 0; }
        [Tool(3, indexCreateTool, Width = 40)]
        public ValueWatcher<int> LoopCount { get { return _loop; } }
        private readonly ValueWatcher<int> _loop = new ValueWatcherFunc<int>(0, i => i.LimitMin(0));

        #endregion

        #region Visual
        const int __cell_color = 1;

        [Menu(1, __cell_color, "CellColor", indexVisual)]
        public void Background()
        {
            var color = this.ControlBackColor;
            this.ControlBackColor = WindowForms.ColorBrowser()
                .ShowDialogSelectOrDefault(color, color, this.PageInfo.Window);
        }

        [Menu(2, __cell_color, indexVisual)]
        public void Text()
        {
            var color = this.CellForeColor;
            this.CellForeColor = WindowForms.ColorBrowser()
                .ShowDialogSelectOrDefault(color, color, this.PageInfo.Window);
        }

        [Menu(3, __cell_color, indexVisual)]
        public void Selected()
        {
            var color = this.CellSelectionBackColor;
            this.CellSelectionBackColor = WindowForms.ColorBrowser()
                .ShowDialogSelectOrDefault(color, color, this.PageInfo.Window);
        }

        [Menu(4, __cell_color, indexVisual)]
        public void SelectedText()
        {
            var color = this.CellSelectionForeColor;
            this.CellSelectionForeColor = WindowForms.ColorBrowser()
                .ShowDialogSelectOrDefault(color, color, this.PageInfo.Window);
        }

        [Menu(8, indexVisual)]
        public void BackColor()
        {
            var color = this.Content.BackColor;
            this.Content.BackColor = WindowForms.ColorBrowser()
                .ShowDialogSelectOrDefault(color, color, this.PageInfo.Window);
        }

        [Menu(9, indexVisual)]
        public void Font()
        {
            var fd = new FontDialog();
            var form = this.Content.FindForm();
            fd.Font = this.Content.Font;

            var result = fd.ShowDialog(form);
            if (result == DialogResult.OK)
            {
                form.Font = fd.Font;
                foreach (var c in form.Controls(true))
                {
                    c.Font = fd.Font;
                }
                foreach (var c in form.StripItems(true))
                {
                    c.Font = fd.Font;
                }
            }
        }

        [Menu(1 + indexInnerCoreLast, indexVisual, ValueExportable = true)]
        public bool DataIndexBar
        {
            get { return this.Content.IsIndexControlVisible; }
            set { this.Content.IsIndexControlVisible = value; }
        }

        [ApplicationSetting]
        public Color ControlBackColor
        {
            get { return this.Content.BackColor; }
            set { this.Content.BackColor = value; }
        }
        [ApplicationSetting]
        public Color CellBackColor
        {
            get { return this.Content.CellBackColor; }
            set { this.Content.CellBackColor = value; }
        }
        [ApplicationSetting]
        public Color CellForeColor
        {
            get { return this.Content.CellForeColor; }
            set { this.Content.CellForeColor = value; }
        }
        [ApplicationSetting]
        public Color CellSelectionBackColor
        {
            get { return this.Content.CellSelectionBackColor; }
            set { this.Content.CellSelectionBackColor = value; }
        }
        [ApplicationSetting]
        public Color CellSelectionForeColor
        {
            get { return this.Content.CellSelectionForeColor; }
            set { this.Content.CellSelectionForeColor = value; }
        }

        #endregion

        #region Status

        [Status(8)]
        public void IndexRangeChanged(ToolStripStatusLabel label)
        {
            Assert.NotNull(label);
            if (label == null)
                return;

            Action<Control> action = _ => label.Text = "-";
            this.PageInfo.SettingImporting += action;
            this.PageInfo.SettingExporting += action;
            this.PageInfo.SettingImported += menu => label.Text = "Setting Imported";
            this.PageInfo.SettingExported += menu => label.Text = "Setting Exported";
            this.Content.IndexRangeChanged += () =>
            {
                var range = this.Content.IndexRange;
                label.Text = "Size: " + range.Width + " x " + range.Height;
            };

            this.PageInfo.Loaded += () => label.Text = "DRAG and DROP IMAGES";
        }

        [Status(7)]
        public void SelectionIndexRangeChanged(ToolStripStatusLabel label)
        {
            Assert.NotNull(label);
            if (label == null)
                return;

            Action<Control> action = _ => label.Text = "-";
            this.PageInfo.SettingImporting += action;
            this.PageInfo.SettingExporting += action;
            this.PageInfo.SettingImported += action;
            this.PageInfo.SettingExported += action;
            this.Content.SelectionIndexRangeChanged += (_, __) =>
            {
                var range = this.Content.SelectionIndexRange;
                label.Text = "| "
                    + (
                        (range.X + ", " + range.Y)
                        + ((range.Width * range.Height <= 1) ? "" : " (" + range.Width + " x " + range.Height + ")")
                    )
                    + " Selected";
            };

            this.PageInfo.Loaded += () => label.Text = "!!!";
        }

        #endregion

        #endregion

        private DirectoryInfo _root_dir = Files.DeskTop;
        private readonly GifSegments _data;
        private readonly DragDropFiles _drop_files = new DragDropFiles();
        public MovinGifMenuSet(PageInfo page)
            : base(page)
        {
            base.Content.ContextMenuStrip = page.GetContextMenu(indexContextMain);

            var list =
#if DEBUG
 Enumerable.Range(0, 4)
                .Select(i => (new GifSegment() { Image = Logo.Image, Interval = i * 10 }))
                .ToList();
#else
 new GifSegment[0];
#endif

            var selectors = TypeInference.Array
            (
            #region Image
GifSegments
                .CreateSelector("Image",
                gs => gs.Image.GetImage(),
                (gs, img) =>
                {
                    gs.Image = img as Image;
                    gs.File = null;
                    return gs;
                }, typeof(Image)),
            #endregion
            #region File
 GifSegments
                .CreateSelector("File",
                gs =>
                {
                    var file = gs.File;
                    if (file == null)
                        return "";

                    var path = _root_dir.GetRelativePath(file);
                    return path;
                },
                (gs, path) =>
                {
                    var file = _root_dir.GetRelativeFile(path.ToString());
                    if (file.Exists)
                    {
                        try
                        {
                            var img = Image.FromFile(file.FullName);
                            gs.Image = img;
                            gs.File = file;
                        }
                        catch (Exception) { }
                    }
                    return gs;
                }, typeof(string)),
            #endregion
            #region Interval
 GifSegments
                .CreateSelector("Interval",
                gs => gs.Interval,
                (gs, itv) =>
                {
                    if (itv is int)
                    {
                        gs.Interval = (int)itv;
                    }
                    else if (itv != null)
                    {
                        gs.Interval = itv.ToString().To(gs.Interval);
                    }

                    gs.Interval = (gs.Interval.LimitMin(10) / 10) * 10;
                    return gs;
                }, typeof(int)),
            #endregion
            #region Optimization
 GifSegments
                .CreateSelector("Optimization",
                gs => gs.DisposalMothod, (gs, value) =>
                {
                    if (value != null)
                    {
                        if (value.GetType() == typeof(GifData.DisposalMethod))
                        {
                            gs.DisposalMothod = (GifData.DisposalMethod)value;
                        }
                        else
                        {
                            gs.DisposalMothod = value.ToString().To(GifData.DisposalMethod.None);
                        }
                    }
                    return gs;
                }, typeof(GifData.DisposalMethod)),
            #endregion
            #region LocalColorTable
 GifSegments
                .CreateSelector("LocalColorTable",
                gs => gs.LocalColorTable, (gs, value) =>
                {
                    if (value != null)
                    {
                        if (value.GetType() == typeof(bool))
                        {
                            gs.LocalColorTable = (bool)value;
                        }
                        else
                        {
                            gs.LocalColorTable = value.ToString().To(false);
                        }
                    }
                    return gs;
                }, typeof(bool))
            #endregion
);
            _data = new GifSegments(list, selectors, false);
            base.Content.Items = _data;
            base.Content.RowHeight = 40;
            base.Content.ItemEdited += (c, r) =>
            {
                if (c <= 1)
                {
                    base.Content.RefleshAllCell();
                }
            };

            this.Optimization.ValueChanged += () =>
            {
                if (this.Optimization.Value)
                {
                    this.SetAllDisposalMethods();
                }
            };

            _drop_files.SetDragDrop(this.Content);
            _drop_files.DragDroped = this.SetFiles;
        }
        private void SetFiles(FileInfo[] files)
        {
            var io = Files.Image();
            foreach (var f in files)
            {
                var img = io.Load(f, FileOpenOptions.Load);
                if (img == null)
                    continue;

                if (f.Extension.ToLower() == ".gif")
                {
                    this.FromGifImage(f, img, plus =>
                    {
                        _data.Records = _data.Records.Concat(plus).ToArray();
                        this.Content.RefleshAllCell();
                    });
                }
                else
                {
                    var plus = TypeInference.Array(new GifSegment()
                    {
                        Image = img,
                        File = f,
                        Interval = 100,
                        DisposalMothod = GifData.DisposalMethod.Fill
                    });
                    _data.Records = _data.Records.Concat(plus).ToArray();
                    this.Content.RefleshAllCell();
                }
            }
        }
        private void FromGifImage(FileInfo file, ImageData image, Action<GifSegment[]> callback)
        {
            this.FromGifImage(TypeInference.Array(TypeInference.KeyValuePair(file, image)), callback);
        }
        private void FromGifImage(IEnumerable<KeyValuePair<FileInfo, ImageData>> data, Action<GifSegment[]> callback)
        {
            int steps = data.Sum(pair => pair.Value.SourceIsAnimationGif ? (pair.Value.GetGifData() as GifData.Animation).Frames.Count : 0);
            if (steps == 0)
            {
                callback(data.Select(pair => new GifSegment()
                {
                    Image = pair.Value,
                    File = pair.Key,
                    DisposalMothod = GifData.DisposalMethod.Fill,
                    Interval = 100,
                    LocalColorTable = false,
                })
                .ToArray());
            }
            else
            {
                ProgressControls.BeginHeavyComputing(pc =>
                {
                    pc.Start("import frames", steps);
                    var result = data
                        .SelectMany(pair =>
                        {
                            var file = pair.Key;
                            var image = pair.Value;
                            var anm = image.SourceIsAnimationGif ? image.GetGifData() as GifData.Animation : null;
                            if (anm != null)
                            {
                                return anm
                                    .ToSource(anm.Frames.Count == steps ? pc : pc.InnerProgress(anm.Frames.Count))
                                    .Select(frame => new GifSegment()
                                    {
                                        Image = frame.Image,
                                        File = file,
                                        DisposalMothod = frame.Optimization,
                                        Interval = frame.DelayTimeCentSec * 10,
                                        LocalColorTable = anm.GlobalColorTable != null && !object.ReferenceEquals(anm.GlobalColorTable, frame.Image.ColorIndexTable)
                                    });
                            }
                            else
                            {
                                return TypeInference.Array(new GifSegment()
                                {
                                    Image = image,
                                    File = file,
                                    DisposalMothod = GifData.DisposalMethod.Fill,
                                    Interval = 100,
                                    LocalColorTable = false,
                                });
                            }
                        })
                        .ToArray();
                    return result;
                }, this.PageInfo.Window, callback);
            }
        }
        private GifData.Animation Create(ProgressObserver pc = null)
        {
            Color? background = null;
            if (this.UseBackColor.Value)
                background = this.Content.BackColor;

            var args = this.CommonColorPalette.Value ? new ColorTableMergeArguments() : null;
            try
            {
                var gif = Create(_data, this.LoopCount.Value, background,
                    this.Transparent.Value, this.CommonColorPalette.Value, this.Optimization.Value, args, pc);
                return gif;
            }
            catch (Exception e)
            {
                this.Message.Text = e.Message;
                throw;
            }
        }
        public static GifData.Animation Create(GifSegments data,
            int loop, Color? background, bool transparent, bool commonTable, bool optimize,
            ColorTableMergeArguments commonColorTable, ProgressObserver progress)
        {
            var images = data
                .Records
                .Select(item => item.Image);
            var local_table_frames = data.Records.Indexed()
                .Where(i => i.Element.LocalColorTable)
                .SelectIndex()
                .ToArray();
            var source = GifAnimationSource.Create(images, commonColorTable, progress, local_table_frames);
            Assert.True(source.Count == data.AsGrid.RowCount);

            foreach (var item_pair in data.Records.Indexed())
            {
                var item = item_pair.Element;
                var s = source[item_pair.Index];

                s.DelayTimeCentSec = item.Interval / 10;
                s.Optimization = item.DisposalMothod;
            }

            if (optimize || background.HasValue || commonTable)
            {
                source.ColorTableOptimize(background, optimize, optimize, commonTable ? (commonColorTable ?? 0x20) : null, local_table_frames);
            }

            if (transparent && background.HasValue && background.Value.A != 0x0)
            {
                var trans_color = Color.FromArgb(background.Value.ToArgb() & 0xFFFFFF);
                foreach (var frame in source)
                {
                    var trans = frame.ColorIndexTable.Add(trans_color, true);
                    var from = frame.ColorIndexTable.IndexOf(background.Value);
                    if (trans >= 0 && from >= 0)
                    {
                        frame.Image.Filter((x, y, index) => (index == from) ? trans : index);
                    }
                }
            }

            if (optimize)
                source.DisposalOptimize();

            var result = source.ToGifData();
            result.LoopCount = loop;
            return result;
        }

        protected override string VersionNameImpl() { return "公開されたコードによるビルド"; }
        public override string ToString() { return "MovinGif"; }
    }
    public struct GifSegment
    {
        public ImageData Image { get; set; }
        public FileInfo File { get; set; }
        public GifData.DisposalMethod DisposalMothod { get; set; }
        public int Interval { get; set; }
        public bool LocalColorTable { get; set; }
    }
}
